/**
 * @class EventHandler
 * @name core.EventHandler
 * @description The EventHandler class is used to handle events in the DocumentViewer control.
 **/
class EventHandler {
  constructor() {
    this._listeners = Object.create(null);
  }

  /**
   * @function
   * @name core.EventHandler#addEvent
   * @param {string} eventName     - The name of the event to listen for.
   * @param {function} listener    - The function to call when the event is dispatched.
   * @param {object} [options]     - The options for the event listener.
   * @param {boolean} options.once - Whether the listener should only be called once.
   * @inner
   */
  addEvent(eventName, listener, options = null) {
    this._addEvent(eventName, listener, {
      external: true,
      once: options?.once,
    });
  }

  /**
   * @function
   * @name core.EventHandler#removeEvent
   * @param {string} eventName  - The name of the event to remove listening for.
   * @param {function} listener - The function to remove from the event listeners.
   */
  removeEvent(eventName, listener) {
    this._removeEvent(eventName, listener, {
      external: true,
      once: options?.once,
    });
  }

  /**
   * @function
   * @name core.EventHandler#dispatch
   * @param {string} eventName - The name of the event to dispatch.
   * @param {*} data           - The data to pass to the event listeners.
   */
  dispatch(eventName, data) {
    const eventListeners = this._listeners[eventName];
    if (!eventListeners || eventListeners.length === 0) {
      return;
    }
    let externalListeners;
    // Making copy of the listeners array in case if it will be modified
    // during dispatch.
    for (const { listener, external, once } of eventListeners.slice(0)) {
      if (once) {
        this._removeEvent(eventName, listener);
      }
      if (external) {
        (externalListeners ||= []).push(listener);
        continue;
      }
      listener(data);
    }
    // Dispatch any "external" listeners *after* the internal ones, to give the
    // viewer components time to handle events and update their state first.
    if (externalListeners) {
      for (const listener of externalListeners) {
        listener(data);
      }
      externalListeners = null;
    }
  }

  /**
   * @ignore
   */
  _addEvent(eventName, listener, options = null) {
    const eventListeners = (this._listeners[eventName] ||= []);
    eventListeners.push({
      listener,
      external: options?.external === true,
      once: options?.once === true,
    });
  }

  /**
   * @ignore
   */
  _removeEvent(eventName, listener) {
    const eventListeners = this._listeners[eventName];
    if (!eventListeners) {
      return;
    }
    for (let i = 0, ii = eventListeners.length; i < ii; i++) {
      if (eventListeners[i].listener === listener) {
        eventListeners.splice(i, 1);
        return;
      }
    }
  }
}
